"""This single file is needed to build the Client development dashboard."""

from promptmodel import Client
from main import client as main_client
# Example imports
# from <dirname> import < objectname>

app = Client()

# Example usage
# This is needed to integrate your codebase with the prompt engineering dashboard
app.include(main_client)

app.register_sample(
	name="function_call_test/1",
	content={"user_message" : "What is the weather like in Boston?"}
)

from main import get_current_weather
get_current_weather_desc = {
	"name": "get_current_weather",
	"description": "Get the current weather in a given location",
	"parameters": {
	"type": "object",
	"properties": {
		"location": {
		"type": "string",
		"description": "The city and state, e.g. San Francisco, CA"
		},
		"unit": {
		"type": "string",
		"enum": ["celsius", "fahrenheit"]
		}
	},
	"required": ["location"]
	}
}

app.register_function(
	get_current_weather_desc,
 	get_current_weather
)


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
import asyncio

from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.utils.enums import LocalTask


@pytest.mark.asyncio
async def test_connect_to_gateway(mocker, websocket_client : DevWebsocketClient, mock_websocket: AsyncMock):
    project_uuid = "test_uuid"
    dev_branch_name = "test_branch"
    cli_access_header = {"Authorization": "Bearer testtoken"}
    
    with patch.object(websocket_client, "_DevWebsocketClient__handle_message", new_callable=AsyncMock) as mock_function:
        mock_function.side_effect = ConnectionClosedOK(None, None)
        with patch("promptmodel.websocket.websocket_client.connect", new_callable=MagicMock, return_value=mock_websocket) as mock_connect:
            # 5초 후에 자동으로 테스트 종료
            await websocket_client.connect_to_gateway(project_uuid, dev_branch_name, cli_access_header, retries=1),
            mock_connect.assert_called_once()
            mock_websocket.recv.assert_called_once()
            websocket_client._DevWebsocketClient__handle_message.assert_called_once()

@pytest.mark.asyncio
async def test_local_tasks(mocker, websocket_client : DevWebsocketClient, mock_websocket: AsyncMock):
    websocket_client._client.llm_modules = {}
    websocket_client._client.samples = {}
    websocket_client._client.functions = {
        "test_function": "test_function"
    }
    
    await websocket_client._DevWebsocketClient__handle_message(
        message = {
            "type" : LocalTask.LIST_FUNCTIONS
        },
        ws = mock_websocket
    )

    
    


# test.py
from promptmodel import PromptModel, Client

client = Client()

prompt = PromptModel("function_call_test").get_prompts()

# prompt = PromptModel("function_call_test").get_prompts()

from typing import Optional
def get_current_weather(location: str, unit: Optional[str]):
    return "13 degrees celsius"

    
# @client.register_agent("lawyer_bot")
# def lawyer_bot(user_message: str):
#     context = client.get_agent_context(agent_name="lawyer_bot")
    
#     res = PromptModel("func_1", use_in_agent=True, tag="agent").run() # 각 run 하나하나가 Agent 내에서 설정된 version을 가져온다
    
#     res = PromptModel("func_1", context = context).run() 
#     # 데코레이터 안에서 flag를 가지고 있거나, 데코레이터가 agent에 대해 설정된 모든 version을 미리 가져오고,
#     # PromptModel은 이게 있는지 확인해서 이를 사용한다


import os
import sys
import importlib
from typing import Any, Dict, List
from threading import Timer
from rich import print
from watchdog.events import FileSystemEventHandler

from promptmodel.apis.base import APIClient
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils import logger
from promptmodel import Client
from promptmodel.database.crud import (
    list_llm_modules,
    update_local_usage_llm_module_by_name,
    create_llm_modules,
    create_llm_module_versions,
    create_prompts,
    create_run_logs,
    update_samples,
    get_llm_module_uuid,
    update_llm_module_uuid,
)
from promptmodel.utils.enums import (
    LLMModuleVersionStatus,
    ChangeLogAction,
)
from promptmodel.websocket.websocket_client import DevWebsocketClient


class CodeReloadHandler(FileSystemEventHandler):
    def __init__(
        self,
        _client_filename: str,
        _instance_name: str,
        dev_websocket_client: DevWebsocketClient,
    ):
        self._client_filename: str = _client_filename
        self.client_instance_name: str = _instance_name
        self.dev_websocket_client: DevWebsocketClient = dev_websocket_client  # 저장
        self.timer = None

    def on_modified(self, event):
        """Called when a file or directory is modified."""
        if event.src_path.endswith(".py"):
            if self.timer:
                self.timer.cancel()
            # reload modified file & main file
            self.timer = Timer(0.5, self.reload_code, args=(event.src_path,))
            self.timer.start()

    def reload_code(self, modified_file_path: str):
        print(
            f"[violet]promptmodel:dev:[/violet]  Reloading {self._client_filename} module due to changes..."
        )
        relative_modified_path = os.path.relpath(modified_file_path, os.getcwd())
        # Reload the client module
        module_name = relative_modified_path.replace("./", "").replace("/", ".")[
            :-3
        ]  # assuming the file is in the PYTHONPATH

        if module_name in sys.modules:
            module = sys.modules[module_name]
            importlib.reload(module)

        reloaded_module = importlib.reload(sys.modules[self._client_filename])
        print(
            f"[violet]promptmodel:dev:[/violet]  {self._client_filename} module reloaded successfully."
        )

        new_client_instance: Client = getattr(
            reloaded_module, self.client_instance_name
        )
        # print(new_client_instance.llm_modules)
        new_llm_module_name_list = [
            llm_module.name for llm_module in new_client_instance.llm_modules
        ]
        old_llm_module_name_list = [
            llm_module.name
            for llm_module in self.dev_websocket_client._client.llm_modules
        ]

        # 사라진 llm_modules 에 대해 local db llm_module.local_usage False Update
        removed_name_list = list(
            set(old_llm_module_name_list) - set(new_llm_module_name_list)
        )
        for removed_name in removed_name_list:
            update_local_usage_llm_module_by_name(removed_name, False)

        # 새로 생긴 llm_module 에 대해 local db llm_module.local_usage True Update
        # TODO: 좀 더 specific 한 API와 연결 필요
        config = read_config()
        org = config["dev_branch"]["org"]
        project = config["dev_branch"]["project"]
        project_status = APIClient.execute(
            method="GET",
            path="/pull_project",
            params={"project_uuid": project["uuid"]},
        ).json()

        changelogs = APIClient.execute(
            method="GET",
            path="/get_changelog",
            params={
                "project_uuid": project["uuid"],
                "local_project_version": config["dev_branch"]["project_version"],
                "levels": [1, 2],
            },
        ).json()
        # IF local_usage=False 인 name=name 이 있을 경우, local_usage=True
        update_by_changelog_for_reload(
            changelogs=changelogs,
            project_status=project_status,
            local_code_llm_module_name_list=new_llm_module_name_list,
        )

        for llm_module in new_client_instance.llm_modules:
            if llm_module.name not in old_llm_module_name_list:
                update_local_usage_llm_module_by_name(llm_module.name, True)

        # create llm_modules in local DB
        db_llm_module_list = list_llm_modules()
        db_llm_module_name_list = [x["name"] for x in db_llm_module_list]
        only_in_local_names = list(
            set(new_llm_module_name_list) - set(db_llm_module_name_list)
        )
        only_in_local_llm_modules = [
            {"name": x, "project_uuid": project["uuid"]} for x in only_in_local_names
        ]
        create_llm_modules(only_in_local_llm_modules)

        # update samples in local DB
        update_samples(new_client_instance.samples)
        self.dev_websocket_client.update_client_instance(new_client_instance)


def update_by_changelog_for_reload(
    changelogs: list[dict],
    project_status: dict,
    local_code_llm_module_name_list: list[str],
):
    """Update Local DB by changelog"""
    local_db_llm_module_list: list = list_llm_modules()  # {"name", "uuid"}

    for changelog in changelogs:
        level: int = changelog["level"]
        logs = changelog['logs']
        if level == 1:
            for log in logs:
                subject = log['subject']
                action:str = log['action']
                if subject == "llm_module":
                    local_db_llm_module_list = update_llm_module_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log['identifiers'],
                        local_db_llm_module_list=local_db_llm_module_list,
                        local_code_llm_module_name_list=local_code_llm_module_name_list
                    )
                elif subject == "llm_module_version":
                    local_db_llm_module_list = update_llm_module_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log['identifiers'],
                        local_db_llm_module_list=local_db_llm_module_list,
                        local_code_llm_module_name_list=local_code_llm_module_name_list
                    )
                else:
                    pass
            previous_version_levels = changelog['previous_version'].split(".")
            current_version_levels = [str(int(previous_version_levels[0]) + 1), "0", "0"]
            current_version = ".".join(current_version_levels)
        elif level == 2:
            for log in logs:
                subject = log['subject']
                action:str = log['action']
                uuid_list:list = log['identifiers']
                if subject == "llm_module_version":
                    local_db_llm_module_list = update_llm_module_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log['identifiers'],
                        local_db_llm_module_list=local_db_llm_module_list,
                        local_code_llm_module_name_list=local_code_llm_module_name_list
                    )
                else:
                    pass
            previous_version_levels = changelog['previous_version'].split(".")
            current_version_levels = [previous_version_levels[0], str(int(previous_version_levels[1]) + 1), "0"]
            current_version = ".".join(current_version_levels)
        else:
            previous_version_levels = changelog['previous_version'].split(".")
            current_version_levels = [previous_version_levels[0], previous_version_levels[1], str(int(previous_version_levels[2]) + 1)]
            current_version = ".".join(current_version_levels)

        upsert_config({"project_version": current_version}, section="dev_branch")
    return True

def update_llm_module_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: list[str],
    local_db_llm_module_list : list[dict],
    local_code_llm_module_name_list : list[str]
):
    if action == ChangeLogAction.ADD.value:
        llm_module_list = [x for x in project_status["llm_modules"] if x["uuid"] in uuid_list]
        for llm_module in llm_module_list:
            local_db_llm_module_name_list = [x["name"] for x in local_db_llm_module_list]

            if llm_module["name"] not in local_db_llm_module_name_list:
                # IF llm_module not in Local DB
                if llm_module["name"] in local_code_llm_module_name_list:
                    # IF llm_module in Local Code
                    llm_module["local_usage"] = True
                    llm_module["is_deployment"] = True
                    create_llm_modules([llm_module])
                else:
                    llm_module["local_usage"] = False
                    llm_module["is_deployment"] = True
                    create_llm_modules([llm_module])
            else:
                # Fix UUID of llm_module
                local_uuid = get_llm_module_uuid(llm_module["name"])['uuid']
                update_llm_module_uuid(local_uuid, llm_module["uuid"])
                local_db_llm_module_list : list = list_llm_modules() 
    else:
        # TODO: add code DELETE, CHANGE, FIX later
        pass
    
    return local_db_llm_module_list
                    
def update_llm_module_version_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: list[str],
    local_db_llm_module_list : list[dict],
    local_code_llm_module_name_list : list[str]
) -> List[Dict[str, Any]]:
    if action == ChangeLogAction.ADD.value:            
       # find llm_module_version in project_status['llm_module_versions'] where uuid in uuid_list
        llm_module_version_list_to_update = [x for x in project_status['llm_module_versions'] if x['uuid'] in uuid_list]
        # check if llm_module_version['name'] is in local_code_llm_module_list
        
        # find prompts and run_logs to update
        prompts_to_update = [x for x in project_status['prompts'] if x['version_uuid'] in uuid_list]
        run_logs_to_update = [x for x in project_status['run_logs'] if x['version_uuid'] in uuid_list]
        
        for llm_module_version in llm_module_version_list_to_update:
            llm_module_version['candidate_version'] = llm_module_version['version']
            del llm_module_version['version']
            llm_module_version['status'] = LLMModuleVersionStatus.CANDIDATE.value
            
        create_llm_module_versions(llm_module_version_list_to_update)
        create_prompts(prompts_to_update)
        create_run_logs(run_logs_to_update)
        
        # local_db_llm_module_list += [{"name" : x['name'], "uuid" : x['uuid']} for x in llm_module_version_list_to_update]
        return local_db_llm_module_list
    else:
        pass




import pytest
import re 
from datetime import datetime
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.types import LLMResponse, LLMStreamResponse
from promptmodel.utils.enums import ParsingPattern, ParsingType, get_pattern_by_type

def generator_format(response: str):
    return {
    "choices" : [
        {
            "delta" : {
                "content" : response
            },
            "finish_reason" : None
        }
    ]
}

def string_to_generator(input_string: str):
    def generator():
        for char in input_string:
            yield generator_format(char)
    return generator()

import random

def random_chunk_generator(input_string: str):
    def generator():
        start = 0
        while start < len(input_string):
            chunk_size = random.randint(1, 3)
            end = min(start + chunk_size, len(input_string))
            yield generator_format(input_string[start:end])
            start = end
    return generator()

async def string_to_agenerator(input_string: str):
    async def agenerator():
        for char in input_string:
            yield generator_format(char)
    return agenerator()

import random

async def random_chunk_agenerator(input_string: str):
    async def agenerator():
        start = 0
        while start < len(input_string):
            chunk_size = random.randint(1, 3)
            end = min(start + chunk_size, len(input_string))
            yield generator_format(input_string[start:end])
            start = end
    return agenerator()

class ResultClass:
    raw_output : str
    parsed_outputs : Dict[str, str]
    parse_success : bool
    
    def __init__(self, raw_output, parsed_outputs, parse_success):
        self.raw_output = raw_output
        self.parsed_outputs = parsed_outputs
        self.parse_success = parse_success
        
bracket_allow_example_gen = string_to_generator("[key type=str]ab[/key]")
bracket_allow_result = ResultClass("[key type=str]ab[/key]", {"key" : "ab"}, True)

bracket_forbidden_example_gen = string_to_generator("[key type=str]ab[/not key]")
bracket_forbidden_result = ResultClass("[key type=str]ab[/not key]", {}, False)
bracket_forbidden_stream_result = ResultClass("[key type=str]ab[/not key]", {"key" : "ab[/not key]"}, False)

double_allow_example_gen = string_to_generator("[[key type=str]]abc[not key]xy[/not key]de[[/key]]")
double_allow_result = ResultClass("[[key type=str]]abc[not key]xy[/not key]de[[/key]]", {"key" : "abc[not key]xy[/not key]de"}, True)

double_forbidden_example_gen = string_to_generator("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]")
double_forbidden_result = ResultClass("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]", {}, False)
double_forbidden_stream_result = ResultClass("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]", {"key" : "abc[not key]xy[/not key]de"}, False)


def generator_processor(generator: Generator):
    raw_output = ""
    parsed_outputs = {}
    parse_success = True
    for chunk in generator:
        # print(chunk)
        if type(chunk) == str:
            raw_output += chunk
        elif type(chunk) == dict:
            for key, value in chunk.items():
                if key in parsed_outputs:
                    parsed_outputs[key] += value
                else:
                    parsed_outputs[key] = value
        elif type(chunk) == bool:
            parse_success = chunk
        elif type(chunk) == LLMResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
        elif type(chunk) == LLMStreamResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
            
    return raw_output, parsed_outputs, parse_success


async def agenerator_processor(generator: AsyncGenerator):
    raw_output = ""
    parsed_outputs = {}
    parse_success = True
    async for chunk in generator:
        # print(chunk)
        if type(chunk) == str:
            raw_output += chunk
        elif type(chunk) == dict:
            for key, value in chunk.items():
                if key in parsed_outputs:
                    parsed_outputs[key] += value
                else:
                    parsed_outputs[key] = value
        elif type(chunk) == bool:
            parse_success = chunk
        elif type(chunk) == LLMResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
        elif type(chunk) == LLMStreamResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
            
    return raw_output, parsed_outputs, parse_success

def test_sp_generator(mocker):
    llm = LLM()
    
    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__single_type_sp_generator__(None, bracket_allow_example_gen, ParsingType.SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)

    assert raw_output == bracket_allow_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == bracket_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == bracket_allow_result.parse_success, f"parse success mismatch : {parse_success}"
    
    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__single_type_sp_generator__(None, bracket_forbidden_example_gen, ParsingType.SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)
    assert raw_output == bracket_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == bracket_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == bracket_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"
    # test Random Generator 5 times 
    for i in range(5):
        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__single_type_sp_generator__(None, random_chunk_generator("[key type=str]ab[/key]"), ParsingType.SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)

        assert raw_output == bracket_allow_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == bracket_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == bracket_allow_result.parse_success, f"parse success mismatch : {parse_success}"
    
        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__single_type_sp_generator__(None, random_chunk_generator("[key type=str]ab[/not key]"), ParsingType.SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)
        assert raw_output == bracket_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == bracket_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == bracket_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"
 
def test_dp_generator(mocker):
    llm = LLM()
    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__double_type_sp_generator__(None, double_allow_example_gen, ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)

    assert raw_output == double_allow_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == double_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == double_allow_result.parse_success, f"parse success mismatch : {parse_success}"


    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__double_type_sp_generator__(None, double_forbidden_example_gen, ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)
    assert raw_output == double_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == double_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == double_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"

    # test random generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__double_type_sp_generator__(None, random_chunk_generator("[[key type=str]]abc[not key]xy[/not key]de[[/key]]"), ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)

        assert raw_output == double_allow_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == double_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == double_allow_result.parse_success, f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__double_type_sp_generator__(None, random_chunk_generator("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"), ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)
        assert raw_output == double_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == double_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == double_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"

@pytest.mark.asyncio
async def test_sp_agenerator(mocker):
    llm = LLM()
    bracket_allow_example_agen = await string_to_agenerator("[key type=str]ab[/key]")
    bracket_forbidden_example_agen = await string_to_agenerator("[key type=str]ab[/not key]")
    
    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__single_type_sp_agenerator__(None, bracket_allow_example_agen, ParsingType.SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)

    assert raw_output == bracket_allow_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == bracket_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == bracket_allow_result.parse_success, f"parse success mismatch : {parse_success}"
    
    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__single_type_sp_agenerator__(None, bracket_forbidden_example_agen, ParsingType.SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)
    assert raw_output == bracket_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == bracket_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == bracket_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"
    # test Random Generator 5 times 
    for i in range(5):
        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__single_type_sp_agenerator__(None, await random_chunk_agenerator("[key type=str]ab[/key]"), ParsingType.SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)

        assert raw_output == bracket_allow_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == bracket_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == bracket_allow_result.parse_success, f"parse success mismatch : {parse_success}"
    
        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__single_type_sp_agenerator__(None, await random_chunk_agenerator("[key type=str]ab[/not key]"), ParsingType.SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)
        assert raw_output == bracket_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == bracket_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == bracket_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"

@pytest.mark.asyncio
async def test_dp_agenerator(mocker):
    llm = LLM()
    double_allow_example_agen = await string_to_agenerator("[[key type=str]]abc[not key]xy[/not key]de[[/key]]")
    double_forbidden_example_agen = await string_to_agenerator("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]")
    
    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__double_type_sp_agenerator__(None, double_allow_example_agen, ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)

    assert raw_output == double_allow_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == double_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == double_allow_result.parse_success, f"parse success mismatch : {parse_success}"


    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__double_type_sp_agenerator__(None, double_forbidden_example_agen, ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
    )
    print(raw_output, parsed_outputs, parse_success)
    assert raw_output == double_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
    assert parsed_outputs == double_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
    assert parse_success == double_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"

    # test random generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__double_type_sp_agenerator__(None, await random_chunk_agenerator("[[key type=str]]abc[not key]xy[/not key]de[[/key]]"), ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)

        assert raw_output == double_allow_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == double_allow_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == double_allow_result.parse_success, f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__double_type_sp_agenerator__(None, await random_chunk_agenerator("[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"), ParsingType.DOUBLE_SQUARE_BRACKET.value, datetime.now())
        )
        print(raw_output, parsed_outputs, parse_success)
        assert raw_output == double_forbidden_stream_result.raw_output, f"raw output error : {raw_output}"
        assert parsed_outputs == double_forbidden_stream_result.parsed_outputs, f"parsed output error : {parsed_outputs}"
        assert parse_success == double_forbidden_stream_result.parse_success, f"parse success mismatch : {parse_success}"
    
@pytest.mark.asyncio
async def test_run_and_parsing(mocker):
    llm = LLM()
    
    mock_response = MagicMock()
    mock_response.choices = [
            {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": "[key type=str]ab[/key]",
                # "function_call": {
                # "name": "get_current_weather",
                # "arguments": "{\n  \"location\": \"Boston, MA\"\n}"
                # }
            },
            "finish_reason": "function_call"
            }
        ]
    mock_api_response = {
        "id": "chatcmpl-8GPN4B6KfJhVqjQK86SAwzvwZe79f",
        "object": "chat.completion",
        "created": 1698921466,
        "model": "gpt-3.5-turbo-0613",
        "choices": [
            {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": "[key type=str]ab[/key]",
                # "function_call": {
                # "name": "get_current_weather",
                # "arguments": "{\n  \"location\": \"Boston, MA\"\n}"
                # }
            },
            "finish_reason": "function_call"
            }
        ],
        "usage": {
            "prompt_tokens": 82,
            "completion_tokens": 18,
            "total_tokens": 100
        }
    }
    mock_completion = mocker.patch("promptmodel.llms.llm.completion", return_value = mock_response)
    
    # success case
    res : LLMResponse = llm.run_and_parse(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"}
        ],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
        functions=[],
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {"key" : "ab"}, f"parsed output error : {res.parsed_outputs}"
    
    # failed case
    res : LLMResponse = llm.run_and_parse(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"}
        ],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
        functions=[],
        output_keys=["key", "key2"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {"key" : "ab"}, f"parsed output error : {res.parsed_outputs}"
    assert res.error == True, f"error mismatch : {res.error}"
    
    # success case with function call
    mock_response.choices[0]["message"]["function_call"] = {
        "name" : "get_current_weather",
        "arguments" : "{\n  \"location\": \"Boston, MA\"\n}"
    }
    res : LLMResponse = llm.run_and_parse(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"}
        ],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
        functions=[],
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {"key" : "ab"}, f"parsed output error : {res.parsed_outputs}"
    assert res.function_call == {
        "name" : "get_current_weather",
        "arguments" : "{\n  \"location\": \"Boston, MA\"\n}"
    }, f"function call mismatch : {res.function_call}"
    
    # error case
    mock_completion = mocker.patch("promptmodel.llms.llm.completion", returun_value = mock_response)
    mocker.patch("promptmodel.llms.llm.LLM.__parse_output_pattern__", side_effect = Exception("test"))
    res : LLMResponse = llm.run_and_parse(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"}
        ],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
        functions=[],
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is not None, f"api response mismatch : {res.api_response}"
    assert res.error == True, f"error mismatch : {res.error}"
    
    mock_completion = mocker.patch("promptmodel.llms.llm.completion", side_effect = Exception("test"), return_value = None)
    res : LLMResponse = llm.run_and_parse(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"}
        ],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
        functions=[],
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is None, f"api response mismatch : {res.api_response}"
    assert res.error == True, f"error mismatch : {res.error}"
    
    
    
    


import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.client import Client

async def echo_coroutine(*args, **kwargs):
    print(args, kwargs)
    return args, kwargs

@pytest.fixture
def websocket_client():
    websocket_client = DevWebsocketClient(
        _client = Client(use_cache=False)
    )
    return websocket_client

@pytest.fixture
def mock_websocket():
    # 모의 WebSocketClientProtocol 객체 생성
    mock_websocket = AsyncMock()
    
    async def aenter(self):
        return self

    async def aexit(self, exc_type, exc_value, traceback):
        pass

    mock_websocket.__aenter__ = aenter
    mock_websocket.__aexit__ = aexit
    mock_websocket.recv = AsyncMock(return_value='{"key" : "value"}')
    mock_websocket.send = AsyncMock(side_effect=echo_coroutine)
    
    return mock_websocket


import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import asyncio
from typing import Optional
from uuid import uuid4
from dataclasses import dataclass
from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.utils.enums import LocalTask, ParsingType
from promptmodel.database.models import LLMModuleVersion
from promptmodel.utils.types import FunctionDescription

@dataclass
class LLMModule:
    name: str
    default_model: str = "gpt-3.5-turbo"

def get_current_weather(location: str, unit: Optional[str] = "celsius"):
        return "13 degrees celsius"
get_current_weather_desc = {
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
    "type": "object",
    "properties": {
        "location": {
        "type": "string",
        "description": "The city and state, e.g. San Francisco, CA"
        },
        "unit": {
        "type": "string",
        "enum": ["celsius", "fahrenheit"]
        }
    },
    "required": ["location"]
    }
}

@pytest.mark.asyncio
async def test_run_model_function_call(mocker, websocket_client : DevWebsocketClient, mock_websocket: AsyncMock):
    websocket_client._client.llm_modules = [LLMModule("test_module")]
    websocket_client._client.samples = {
        "sample_1" : {"user_message" : "What is the weather like in Boston?"}
    }
    websocket_client._client.functions = {
        "get_current_weather": {
            "description" : FunctionDescription(**get_current_weather_desc),
            "function" : get_current_weather
        }
    }
    llm_module_uuid = uuid4()
    param = {
        "llm_module_uuid" : llm_module_uuid,
        "uuid" : uuid4(),
        "status" : "broken",
        "from_uuid" : None,
        "model" : "gpt-3.5-turbo",
        "parsing_type" : None,
        "output_keys" : None,
        "functions" : []
    }
    mock_version = LLMModuleVersion(**param)
    mocker.patch(
        "promptmodel.websocket.websocket_client.get_sample_input",
        new_callable=MagicMock,
        return_value={
            "name" : "sample_1",
            "contents" : {"user_message" : "What is the weather like in Boston?"}
        }
    )
    mocker.patch("promptmodel.websocket.websocket_client.get_llm_module_uuid", new_callable=MagicMock, return_value={"uuid" : llm_module_uuid})
    mocker.patch("promptmodel.websocket.websocket_client.create_llm_module_version", new_callable=MagicMock, return_value=mock_version)
    mocker.patch("promptmodel.websocket.websocket_client.create_prompt", new_callable=MagicMock)
    mocker.patch("promptmodel.websocket.websocket_client.update_llm_module_version", new_callable=MagicMock)
    create_run_log_mock = mocker.patch("promptmodel.websocket.websocket_client.create_run_log", new_callable=MagicMock)
    
    # success case
    await websocket_client._DevWebsocketClient__handle_message(
        message = {
            "type" : LocalTask.RUN_LLM_MODULE,
            "llm_module_name" : "test_module",
            "sample_name" : "sample_1",
            "prompts" : [
                { "role" : "system", "content" : "You are a helpful assistant.", "step" : 1},
                { "role" : "user", "content" : "{user_message}", "step" : 2}
            ],
            "model" : "gpt-3.5-turbo",
            "uuid" : None,
            "from_uuid" : None,
            "parsing_type" : None,
            "output_keys" : None,
            "functions" : ["get_current_weather"]
        },
        ws = mock_websocket
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs['function_call'] is not None, "function_call is None"
    create_run_log_mock.reset_mock()
    print("=======================================================================================")
    
    # success case with no function call
    await websocket_client._DevWebsocketClient__handle_message(
        message = {
            "type" : LocalTask.RUN_LLM_MODULE,
            "llm_module_name" : "test_module",
            "sample_name" : "sample_1",
            "prompts" : [
                { "role" : "system", "content" : "You are a helpful assistant.", "step" : 1},
                { "role" : "user", "content" : "{user_message}", "step" : 2}
            ],
            "model" : "gpt-3.5-turbo",
            "uuid" : None,
            "from_uuid" : None,
            "parsing_type" : None,
            "output_keys" : None,
            "functions" : []
        },
        ws = mock_websocket
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs['function_call'] is None, "function_call is not None"
    create_run_log_mock.reset_mock()
    print("=======================================================================================")
    
    # success case with parsing
    system_prompt_with_format = """
You are a helpful assistant.

This is your output format. Keep the string between < type=< >>, </ > as it is.
<temperature type=<float>>
(value here)
</temperature>
    """
    await websocket_client._DevWebsocketClient__handle_message(
        message = {
            "type" : LocalTask.RUN_LLM_MODULE,
            "llm_module_name" : "test_module",
            "sample_name" : "sample_1",
            "prompts" : [
                { "role" : "system", "content" : system_prompt_with_format, "step" : 1},
                { "role" : "user", "content" : "{user_message}", "step" : 2}
            ],
            "model" : "gpt-3.5-turbo",
            "uuid" : None,
            "from_uuid" : None,
            "parsing_type" : ParsingType.HTML.value,
            "output_keys" : ["temperature"],
            "functions" : ["get_current_weather"]
        },
        ws = mock_websocket
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs['function_call'] is not None, "function_call is None"
    create_run_log_mock.reset_mock()
    print("=======================================================================================")
    
    # early failed case due to function not existing
    await websocket_client._DevWebsocketClient__handle_message(
        message = {
            "type" : LocalTask.RUN_LLM_MODULE,
            "llm_module_name" : "test_module",
            "sample_name" : "sample_1",
            "prompts" : [
                { "role" : "system", "content" : "You are a helpful assistant.", "step" : 1},
                { "role" : "user", "content" : "{user_message}", "step" : 2}
            ],
            "model" : "gpt-3.5-turbo",
            "uuid" : None,
            "from_uuid" : None,
            "parsing_type" : None,
            "output_keys" : None,
            "functions" : ["get_weather"]
        },
        ws = mock_websocket
    )
    create_run_log_mock.assert_not_called()


"""LLM module for Development TestRun"""
import re
import os
import json
import openai

from typing import Any, AsyncGenerator, List, Dict, Optional, Union, Generator
from pydantic import BaseModel
from dotenv import load_dotenv
from litellm import acompletion

from promptmodel.utils.enums import ParsingType, ParsingPattern, get_pattern_by_type
from promptmodel.utils import logger
from promptmodel.utils.output_utils import convert_str_to_type
from promptmodel.utils.types import LLMStreamResponse

load_dotenv()


class OpenAIMessage(BaseModel):
    role: str
    content: str
    
class OpenAIFunctionMessage(BaseModel):
    role: str
    content: str
    name: str


class LLMDev:
    def __init__(self):
        self._model: str

    def __validate_openai_messages(
        self, messages: List[Dict[str, str]]
    ) -> List[Union[OpenAIMessage, OpenAIFunctionMessage]]:
        """Validate and convert list of dictionaries to list of OpenAIMessage."""
        res = []
        for message in messages:
            if "role" in message and message["role"] == "function":
                res.append(OpenAIFunctionMessage(**message))
            else:
                res.append(OpenAIMessage(**message))
        return res

    async def dev_run(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: [List[Any]] = [],
        model: Optional[str] = None,
    ) -> AsyncGenerator[Any, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        response = await acompletion(
            model=_model,
            messages=[
                message.model_dump()
                for message in self.__validate_openai_messages(messages)
            ],
            stream=True,
            functions=functions,
        )
        function_call = {"name" : "", "arguments" : ""}
        finish_reason_function_call = False
        async for chunk in response:
            if "content" in chunk["choices"][0]["delta"] and chunk["choices"][0]["delta"]["content"] is not None:
                stream_value = chunk["choices"][0]["delta"]["content"]
                raw_output += stream_value # append raw output
                yield LLMStreamResponse(raw_output=stream_value) # return raw output
                
            if "function_call" in chunk["choices"][0]["delta"] and chunk["choices"][0]["delta"]["function_call"] is not None:
                for key, value in chunk["choices"][0]["delta"]["function_call"].items():
                    function_call[key] += value

            if chunk["choices"][0]["finish_reason"] == "function_call":
                finish_reason_function_call = True
                yield LLMStreamResponse(function_call=function_call)
                
        # parsing
        if parsing_type and not finish_reason_function_call:
            parsing_pattern: Dict[str, str] = get_pattern_by_type(parsing_type)
            whole_pattern = parsing_pattern["whole"]
            parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                yield LLMStreamResponse(parsed_outputs={key: value})

    async def dev_chat(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        model: Optional[str] = None,
    ) -> AsyncGenerator[Any, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        response = await acompletion(
            model=_model,
            messages=[
                message.model_dump()
                for message in self.__validate_openai_messages(messages)
            ],
            stream=True,
        )
        async for chunk in response:
            if "content" in chunk["choices"][0]["delta"]:
                stream_value = chunk["choices"][0]["delta"]["content"]
                raw_output += stream_value # append raw output
                yield LLMStreamResponse(raw_output=stream_value) # return raw output
                
        # parsing
        if parsing_type:
            parsing_pattern: Dict[str, str] = get_pattern_by_type(parsing_type)
            whole_pattern = parsing_pattern["whole"]
            parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                yield LLMStreamResponse(parsed_outputs={key: value})



from __future__ import annotations
import asyncio
import dis
import json
import inspect
import threading
import time
import atexit
from dataclasses import dataclass
from typing import Callable, Dict, Any, List, Optional, Union
from websockets.client import connect, WebSocketClientProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK

import promptmodel.utils.logger as logger
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.prompt_util import update_deployed_db
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils.types import FunctionDescription
from promptmodel.database.orm import initialize_db


@dataclass
class LLMModule:
    name: str
    default_model: str = "gpt-3.5-turbo"


class Client:
    """Client main class"""

    def __init__(self, use_cache: Optional[bool] = True, default_model: Optional[str] = "gpt-3.5-turbo"):
        self._default_model: str = default_model
        self.llm_modules: List[LLMModule] = []
        self.samples: List[Dict[str, Any]] = []
        self.functions : Dict[str, Dict[str, Union[FunctionDescription, Callable]]] = {}
        config = read_config()
        if (
            config and 
            "dev_branch" in config and (
                (
                    "online" in config['dev_branch'] and config['dev_branch']["online"] == True
                ) or 
                (
                    "initializing" in config['dev_branch'] and config['dev_branch']["initializing"] == True
                )
            )
        ):
            self.cache_manager = None
        else:
            if use_cache:
                upsert_config({"use_cache": True}, section="project")
                self.cache_manager = CacheManager()
            else:
                upsert_config({"use_cache": False}, section="project")
                self.cache_manager = None
                initialize_db() # init db for local usage

    def register(self, func):
        instructions = list(dis.get_instructions(func))
        for idx in range(
            len(instructions) - 1
        ):  # We check up to len-1 because we access idx+1 inside loop
            instruction = instructions[idx]
            # print(instruction)
            if (
                instruction.opname in ["LOAD_ATTR", "LOAD_METHOD", "LOAD_GLOBAL"]
                and instruction.argval == "PromptModel"
            ):
                next_instruction = instructions[idx + 1]

                # Check if the next instruction is LOAD_CONST with string value
                if next_instruction.opname == "LOAD_CONST" and isinstance(
                    next_instruction.argval, str
                ):
                    self.llm_modules.append(
                        LLMModule(
                            name=next_instruction.argval,
                            default_model=self._default_model,
                        )
                    )

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def register_llm_module(self, name):
        for llm_module in self.llm_modules:
            if llm_module.name == name:
                return

        self.llm_modules.append(
            LLMModule(
                name=name,
                default_model=self._default_model,
            )
        )
        
    def register_function(self, description: Union[Dict[str, Any], FunctionDescription], function: Callable):
        function_name = description['name']
        if isinstance(description, dict):
            try:
                description = FunctionDescription(**description)
            except:
                raise ValueError("description is not a valid function call description.")
            
        if function_name not in self.functions:
            self.functions[function_name] = {
                "description" : description,
                "function" : function,
            }

    def include(self, client: Client):
        self.llm_modules.extend(client.llm_modules)
        # delete duplicated llm_modules
        self.llm_modules = list(
            {llm_module.name: llm_module for llm_module in self.llm_modules}.values()
        )

        self.samples.extend(client.samples)
        # delete duplicated samples
        self.samples = list(
            {sample["name"]: sample for sample in self.samples}.values()
        )
        
        self.functions.update(client.functions)

    def register_sample(self, name: str, content: Dict[str, Any]):
        self.samples.append({"name": name, "contents": content})
        
    def _get_llm_module_name_list(self) -> List[str]:
        return [llm_module.name for llm_module in self.llm_modules]
    
    def _call_register_function(self, name: str, arguments: Dict[str, str]):
        function_to_call : Callable = self.functions[name]["function"]
        try:
            function_response = function_to_call(**arguments)
            return function_response
        except Exception as e:
            raise e
        
    def _get_function_name_list(self) -> List[str]:
        return list(self.functions.keys())
    
    def _get_function_descriptions(self, function_names: List[str] = []):
        try:
            function_descriptions = [self.functions[function_name]["description"].model_dump() for function_name in function_names]
            return function_descriptions
        except Exception as e:
            raise e

class DevApp(Client):
    def __init__(self, default_model: Optional[str] = "gpt-3.5-turbo"):
        super().__init__(default_model)

    def include(self, client: Client):
        self.llm_modules.extend(client.llm_modules)


class CacheManager:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(CacheManager, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self.last_update_time = 0  # to manage update frequency
        self.update_interval = 5  # seconds, consider tuning this value
        self.program_alive = True
        initialize_db()
        atexit.register(self._terminate)
        # # logger.debug("CacheManager initialized")
        asyncio.run(self.update_cache()) # updae cache first synchronously
        self.cache_thread = threading.Thread(target=self._run_cache_loop)
        self.cache_thread.daemon = True
        self.cache_thread.start()
        logger.debug("CacheManager initialized")

    def _run_cache_loop(self):
        asyncio.run(self._update_cache_periodically())

    async def _update_cache_periodically(self):
        while True:
            await asyncio.sleep(self.update_interval)  # Non-blocking sleep
            await self.update_cache()

    async def update_cache(self):
        # Current time
        current_time = time.time()
        config = read_config()
        if not config:
            upsert_config({"version": "0.0.0"}, section="project")
            config = {"project" : {"version": "0.0.0"}}

        # Check if we need to update the cache
        if current_time - self.last_update_time > self.update_interval:
            # Update cache logic
            await update_deployed_db(config)
            # Update the last update time
            self.last_update_time = current_time

    def _terminate(self):
        self.program_alive = False



"""This single file is needed to build the Client development dashboard."""

from promptmodel import Client

# Example imports
# from <dirname> import <objectname>

app = Client()

# Example usage
# This is needed to integrate your codebase with the prompt engineering dashboard
# app.include(<objectname>)



import os

from .models import (
    LLMModule,
    LLMModuleVersion,
    Prompt,
    RunLog,
    SampleInputs,
    DeployedLLMModule,
    DeployedLLMModuleVersion,
    DeployedPrompt,
)
from .config import db


def initialize_db():
    if not os.path.exists("./.promptmodel"):
        os.mkdir("./.promptmodel")
    # Check if db connection exists
    if db.is_closed():
        db.connect()
        with db.atomic():
            if not LLMModule.table_exists():
                db.create_tables(
                    [
                        LLMModule,
                        LLMModuleVersion,
                        Prompt,
                        RunLog,
                        SampleInputs,
                        DeployedLLMModule,
                        DeployedLLMModuleVersion,
                        DeployedPrompt,
                    ]
                )
        db.close()



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils.types import LLMResponse, LLMStreamResponse

def test_stream(mocker):
	llm = LLM()
	test_messages = [
		{"role" : "system", "content" : "You are a helpful assistant."},
		{"role" : "user", "content" : "Introduce yourself in 50 words."},
	]
 
	stream_res : Generator[LLMStreamResponse, None, None] = llm.stream(
		messages=test_messages,
		functions=[],
		model="gpt-3.5-turbo",
	)
	error_count = 0
	api_responses : List[ModelResponse] = []
	for res in stream_res:
		if res.error:
			error_count += 1
			print("ERROR")
			print(res.error)
			print(res.error_log)
		if res.api_response:
			api_responses.append(res.api_response)

	assert error_count == 0, "error_count is not 0"
	assert len(api_responses) == 1, "api_count is not 1"
	
	assert api_responses[0].choices[0]['message']['content'] is not None, "content is None"
	assert api_responses[0]['response_ms'] is not None, "response_ms is None"
 
	# test logging
	llm_proxy = LLMProxy("test")
	
	mock_execute = AsyncMock()
	mock_response = MagicMock()
	mock_response.status_code = 200
	mock_execute.return_value = mock_response
	mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)
	mocker.patch("promptmodel.llms.llm_proxy.read_config", new_callable=MagicMock, return_value={})
	
	llm_proxy._log_to_cloud(
		version_uuid="test",
		inputs={},
		api_response=api_responses[0],
		parsed_outputs={},
		metadata={},
	)
 
	mock_execute.assert_called_once()
	_, kwargs = mock_execute.call_args
 
	assert kwargs['json']['api_response'] == api_responses[0].to_dict_recursive(), "api_response is not equal"


import os
import sys
import yaml
import asyncio
from typing import Any, Dict, Tuple, List, Union, Optional

from promptmodel.database.crud import (
    get_latest_version_prompts,
    get_deployed_prompts,
    update_deployed_cache,
)
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils import logger
from promptmodel.apis.base import APIClient, AsyncAPIClient


async def fetch_prompts(name) -> Tuple[List[Dict[str, str]], Dict[str, Any]]:
    """fetch prompts.

    Args:
        name (str): name of promtpmodel

    Returns:
        Tuple[List[Dict[str, str]], Optional[Dict[str, Any]]]: (prompts, version_detail)
    """
    # Check dev_branch activate
    config = read_config()
    if "dev_branch" in config and config["dev_branch"]["initializing"] == True:
        return [], {}
    elif "dev_branch" in config and config["dev_branch"]["online"] == True:
        # get prompt from local DB
        prompt_rows, version_detail = get_latest_version_prompts(name)
        if prompt_rows is None:
            return [], {}
        return [
            {"role": prompt.role, "content": prompt.content}
            for prompt in prompt_rows
        ], version_detail
    else:
        if (
            "project" in config
            and "use_cache" in config["project"]
            and config["project"]["use_cache"] == True
        ):
            # call update_local API in background task
            asyncio.create_task(update_deployed_db(config))
            # get prompt from local DB by ratio
            prompt_rows, version_detail = get_deployed_prompts(name)
            if prompt_rows is None:
                return [], {}
        else:
            await update_deployed_db(config) # wait for update local DB cache
            prompt_rows, version_detail = get_deployed_prompts(name)
            if prompt_rows is None:
                return [], {}

        return [
            {"role": prompt.role, "content": prompt.content} for prompt in prompt_rows
        ], version_detail


async def update_deployed_db(config):
    if "project" not in config or "version" not in config["project"]:
        cached_project_version = "0.0.0"
    else:
        cached_project_version = config["project"]["version"]
    try:
        res = await AsyncAPIClient.execute(
            method="GET",
            path="/check_update",
            params={"cached_version": cached_project_version},
            use_cli_key=False,
        )
        res = res.json()
        if res["need_update"]:
            # update local DB with res['project_status']
            project_status = res["project_status"]
            await update_deployed_cache(project_status)
            upsert_config({"version": res["version"]}, section="project")
        else:
            upsert_config({"version": res["version"]}, section="project")
    except Exception as exception:
        logger.error(f"Deployment cache update error: {exception}")


def set_inputs_to_prompts(inputs: Dict[str, Any], prompts: List[Dict[str, str]]):
    messages = [
        {"content": prompt["content"].format(**inputs), "role": prompt["role"]}
        for prompt in prompts
    ]
    return messages
